<?php

namespace Drupal\existing_filenames_sanitizer\Commands;

use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drush\Commands\DrushCommands;
use Drupal\Component\Transliteration\PhpTransliteration;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\file\Entity\File;

/**
 * Drush commands for Filename Sanitizer module.
 */
class ExistingFilenameSanitizerCommands extends DrushCommands {

  use StringTranslationTrait;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The file system service.
   *
   * @var \Drupal\Core\File\FileSystemInterface
   */
  protected $fileSystem;

  /**
   * The transliteration service.
   *
   * @var \Drupal\Component\Transliteration\PhpTransliteration
   */
  protected $transliteration;

  /**
   * The config factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * Constructs a new FilenameSanitizerCommands object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Core\File\FileSystemInterface $file_system
   *   The file system service.
   * @param \Drupal\Component\Transliteration\PhpTransliteration $transliteration
   *   The transliteration service.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The config factory.
   */
  public function __construct(
    EntityTypeManagerInterface $entity_type_manager,
    FileSystemInterface $file_system,
    PhpTransliteration $transliteration,
    ConfigFactoryInterface $config_factory
  ) {
    parent::__construct();
    $this->entityTypeManager = $entity_type_manager;
    $this->fileSystem = $file_system;
    $this->transliteration = $transliteration;
    $this->configFactory = $config_factory;
  }

  /**
   * Sanitize filenames for existing files.
   *
   * @command existing-filenames-sanitizer:sanitize-filenames
   * @aliases efssf
   * @option dry-run Show what would be changed without making changes
   * @option transliterate Apply transliteration to filenames
   * @option replace-whitespace Replace whitespace with replacement character
   * @option replace-non-alphanumeric Replace non-alphanumeric characters
   * @option deduplicate-separators Remove duplicate separators
   * @option lowercase Convert filenames to lowercase
   * @option replacement-character Character to use for replacements
   * @option skip-temporary Skip temporary files
   * @option skip-missing Skip files where physical file doesn't exist
   * @usage existing-filenames-sanitizer:sanitize-filenames
   *   Sanitize all existing file filenames using current settings
   * @usage existing-filenames-sanitizer:sanitize-filenames --dry-run
   *   Show what would be changed without making changes
   * @usage existing-filenames-sanitizer:sanitize-filenames --transliterate --lowercase --replacement-character=_
   *   Sanitize with transliteration, lowercase, and underscore replacement
   * @usage existing-filenames-sanitizer:sanitize-filenames --skip-temporary
   *   Skip temporary files during sanitization
   * @usage existing-filenames-sanitizer:sanitize-filenames --skip-missing
   *   Skip files where physical file doesn't exist
   */
  public function sanitizeFilenames($options = [
    'dry-run' => FALSE,
    'transliterate' => FALSE,
    'replace-whitespace' => FALSE,
    'replace-non-alphanumeric' => FALSE,
    'deduplicate-separators' => FALSE,
    'lowercase' => FALSE,
    'replacement-character' => '_',
    'skip-temporary' => FALSE,
    'skip-missing' => FALSE,
  ]) {
    $this->output()->writeln($this->t('Starting filename sanitization process...'));

    // Get file settings configuration
    $file_settings = $this->configFactory->get('file.settings');
    $sanitization_settings = $file_settings->get('filename_sanitization');

    // Use command options if provided, otherwise use current settings
    $transliterate = $options['transliterate'] ?: $sanitization_settings['transliterate'];
    $replace_whitespace = $options['replace-whitespace'] ?: $sanitization_settings['replace_whitespace'];
    $replace_non_alphanumeric = $options['replace-non-alphanumeric'] ?: $sanitization_settings['replace_non_alphanumeric'];
    $deduplicate_separators = $options['deduplicate-separators'] ?: $sanitization_settings['deduplicate_separators'];
    $lowercase = $options['lowercase'] ?: $sanitization_settings['lowercase'];
    $replacement_character = $options['replacement-character'] ?: $sanitization_settings['replacement_character'];

    $this->output()->writeln($this->t('Sanitization settings:'));
    $this->output()->writeln($this->t('- Transliterate: @value', ['@value' => $transliterate ? 'Yes' : 'No']));
    $this->output()->writeln($this->t('- Replace whitespace: @value', ['@value' => $replace_whitespace ? 'Yes' : 'No']));
    $this->output()->writeln($this->t('- Replace non-alphanumeric: @value', ['@value' => $replace_non_alphanumeric ? 'Yes' : 'No']));
    $this->output()->writeln($this->t('- Deduplicate separators: @value', ['@value' => $deduplicate_separators ? 'Yes' : 'No']));
    $this->output()->writeln($this->t('- Lowercase: @value', ['@value' => $lowercase ? 'Yes' : 'No']));
    $this->output()->writeln($this->t('- Replacement character: @value', ['@value' => $replacement_character]));

    if ($options['dry-run']) {
      $this->output()->writeln($this->t('DRY RUN MODE - No changes will be made'));
    }

    // Get all file entities
    $storage = $this->entityTypeManager->getStorage('file');
    $query = $storage->getQuery()
      ->condition('status', 1)
      ->accessCheck(FALSE);
    if ($options['skip-temporary']) {
      $query->condition('uri', 'temporary://', 'NOT LIKE');
    }
    $fids = $query->execute();

    if (empty($fids)) {
      $this->output()->writeln($this->t('No permanent files found.'));
      return;
    }

    $this->output()->writeln($this->t('Found @count files to process.', ['@count' => count($fids)]));

    $processed = 0;
    $renamed = 0;
    $errors = 0;

    foreach ($fids as $fid) {
      try {
        $file = $storage->load($fid);
        if (!$file) {
          continue;
        }

        /** @var \Drupal\file\Entity\File $file */
        $original_filename = $file->getFilename();
        
        // Skip temporary files if requested
        if ($options['skip-temporary'] && strpos($file->getFileUri(), 'temporary://') === 0) {
          $this->output()->writeln($this->t('Skipping temporary file @fid: @filename', [
            '@fid' => $fid,
            '@filename' => $original_filename,
          ]));
          continue;
        }
        
        $new_filename = $this->sanitizeFilename(
          $original_filename,
          $transliterate,
          $replace_whitespace,
          $replace_non_alphanumeric,
          $deduplicate_separators,
          $lowercase,
          $replacement_character
        );

        $processed++;

        if ($original_filename !== $new_filename) {
          $this->output()->writeln($this->t('File @fid: "@old" -> "@new"', [
            '@fid' => $fid,
            '@old' => $original_filename,
            '@new' => $new_filename,
          ]));

          if (!$options['dry-run']) {
            $success = $this->renameFile($file, $new_filename, $options['skip-missing']);
            if ($success) {
              $renamed++;
            } else {
              $errors++;
            }
          } else {
            $renamed++;
          }
        }

        // Show progress every 100 files
        if ($processed % 100 === 0) {
          $this->output()->writeln($this->t('Processed @count files...', ['@count' => $processed]));
        }

      } catch (\Exception $e) {
        $this->output()->writeln($this->t('Error processing file @fid: @error', [
          '@fid' => $fid,
          '@error' => $e->getMessage(),
        ]));
        $errors++;
      }
    }

    $this->output()->writeln('');
    $this->output()->writeln($this->t('Sanitization complete:'));
    $this->output()->writeln($this->t('- Total files processed: @count', ['@count' => $processed]));
    $this->output()->writeln($this->t('- Files renamed: @count', ['@count' => $renamed]));
    $this->output()->writeln($this->t('- Errors: @count', ['@count' => $errors]));

    if ($options['dry-run']) {
      $this->output()->writeln($this->t('This was a dry run. Run without --dry-run to apply changes.'));
    }
  }

  /**
   * Sanitize a filename according to the specified rules.
   *
   * @param string $filename
   *   The original filename.
   * @param bool $transliterate
   *   Whether to transliterate.
   * @param bool $replace_whitespace
   *   Whether to replace whitespace.
   * @param bool $replace_non_alphanumeric
   *   Whether to replace non-alphanumeric characters.
   * @param bool $deduplicate_separators
   *   Whether to deduplicate separators.
   * @param bool $lowercase
   *   Whether to convert to lowercase.
   * @param string $replacement_character
   *   The replacement character.
   *
   * @return string
   *   The sanitized filename.
   */
  protected function sanitizeFilename($filename, $transliterate, $replace_whitespace, $replace_non_alphanumeric, $deduplicate_separators, $lowercase, $replacement_character) {
    // Split filename and extension
    $pathinfo = pathinfo($filename);
    $name = $pathinfo['filename'];
    $extension = isset($pathinfo['extension']) ? '.' . $pathinfo['extension'] : '';

    // Apply transliteration
    if ($transliterate) {
      $name = $this->transliteration->transliterate($name);
    }

    // Replace whitespace
    if ($replace_whitespace) {
      $name = preg_replace('/\s+/', $replacement_character, $name);
    }

    // Replace non-alphanumeric characters
    if ($replace_non_alphanumeric) {
      $name = preg_replace('/[^a-zA-Z0-9\-_.]/', $replacement_character, $name);
    }

    // Deduplicate separators
    if ($deduplicate_separators) {
      $name = preg_replace('/' . preg_quote($replacement_character, '/') . '+/', $replacement_character, $name);
      $name = trim($name, $replacement_character);
    }

    // Convert to lowercase
    if ($lowercase) {
      $name = strtolower($name);
    }

    return $name . $extension;
  }

  /**
   * Rename a file entity and its physical file.
   *
   * @param \Drupal\file\Entity\File $file
   *   The file entity.
   * @param string $new_filename
   *   The new filename.
   * @param bool $skip_missing
   *   Whether to skip files where physical file doesn't exist.
   *
   * @return bool
   *   TRUE if successful, FALSE otherwise.
   */
  protected function renameFile($file, $new_filename, $skip_missing = FALSE) {
    try {
      $uri = $file->getFileUri();
      $directory = dirname($uri);
      $pathinfo = pathinfo($new_filename);
      $base = $pathinfo['filename'];
      $ext = isset($pathinfo['extension']) ? '.' . $pathinfo['extension'] : '';
      $unique_filename = $new_filename;
      $new_uri = $directory . '/' . $unique_filename;
      $counter = 1;

      // Check if the file exists
      if (!file_exists($uri)) {
        $this->output()->writeln($this->t('Warning: Physical file does not exist: @uri', ['@uri' => $uri]));

        // For temporary files that don't exist, we can only update the filename in the database
        if (strpos($uri, 'temporary://') === 0) {
          // Find a unique filename for the DB record
          while ($this->fileEntityExists($directory, $unique_filename, $file->id())) {
            $unique_filename = $base . '_' . $counter . $ext;
            $counter++;
          }
          $this->output()->writeln($this->t('File @fid is a temporary file. Only updating database record.', ['@fid' => $file->id()]));
          $file->setFilename($unique_filename);
          $file->save();
          return TRUE;
        }

        // For missing files, check if we should skip them
        if ($skip_missing) {
          $this->output()->writeln($this->t('Skipping file @fid due to missing physical file.', ['@fid' => $file->id()]));
          return FALSE;
        }

        // For oEmbed thumbnails and other generated content, we can still update the database
        if (strpos($uri, 'oembed_thumbnails') !== FALSE || 
            strpos($uri, 'generated') !== FALSE ||
            strpos($uri, 'cache') !== FALSE) {
          // Find a unique filename for the DB record
          while ($this->fileEntityExists($directory, $unique_filename, $file->id())) {
            $unique_filename = $base . '_' . $counter . $ext;
            $counter++;
          }
          $this->output()->writeln($this->t('File @fid appears to be generated content. Only updating database record.', ['@fid' => $file->id()]));
          $file->setFilename($unique_filename);
          $file->save();
          return TRUE;
        }

        return FALSE;
      }

      // Find a unique filename for the physical file
      while (file_exists($new_uri)) {
        $unique_filename = $base . '_' . $counter . $ext;
        $new_uri = $directory . '/' . $unique_filename;
        $counter++;
      }

      // Rename the physical file
      $result = $this->fileSystem->move($uri, $new_uri, FileSystemInterface::EXISTS_RENAME);
      if ($result === FALSE) {
        $this->output()->writeln($this->t('Error: Could not rename physical file: @uri', ['@uri' => $uri]));
        return FALSE;
      }

      // Update the file entity
      $file->setFilename($unique_filename);
      $file->setFileUri($result);
      $file->save();

      return TRUE;

    } catch (\Exception $e) {
      $this->output()->writeln($this->t('Error renaming file @fid: @error', [
        '@fid' => $file->id(),
        '@error' => $e->getMessage(),
      ]));
      return FALSE;
    }
  }

  /**
   * Check if a file entity with the same filename exists in the same directory.
   *
   * @param string $directory
   *   The directory path (e.g., public://2024-06).
   * @param string $filename
   *   The filename to check.
   * @param int $exclude_fid
   *   The file ID to exclude from the check (current file).
   *
   * @return bool
   *   TRUE if a file entity with the same filename exists, FALSE otherwise.
   */
  protected function fileEntityExists($directory, $filename, $exclude_fid = NULL) {
    $storage = $this->entityTypeManager->getStorage('file');
    $query = $storage->getQuery()
      ->condition('uri', $directory . '/' . $filename)
      ->accessCheck(FALSE);
    if ($exclude_fid) {
      $query->condition('fid', $exclude_fid, '!=');
    }
    $fids = $query->execute();
    return !empty($fids);
  }

} 
